# -*- Python -*-

# Having a go at the constraint-based inference algorithm described
#  by Pottier and Remy in ATTPL, chapter 10: "The Essence of ML Type Inference".
#
# Another great reference is a somewhat simplified presentation of the
#  same material, but (thankfully) with some context, by Pottier: "A
#  modern eye on ML type inference - Old techniques and recent
#  developments", available from his home page:
#  http://cristal.inria.fr/~fpottier/
#
# NOTE: this is a work in progress.
# http://www.nightmare.com/rushing/irken/

import sys
is_a = isinstance

import pdb
trace = pdb.set_trace

# the simply typed lambda calculus:
#   e ::= x | \x.e | e e
# 
# expressions:
# x    : <varref>
# \x.e : <function>
# e e  : <application>

# types
# t ::= a | (arrow t t)
# (where a = <tvar>)

# constraints:
# C ::= (equals t t) | (and C C) | (exists a C)

# types

tvar_counter = -1

def fresh():
    global tvar_counter
    result = tvar_counter
    tvar_counter -= 1
    return result

class _type:
    pass

class tvar (_type):
    def __init__ (self):
        self.id = fresh()

    def __repr__ (self):
        #return 'T%d' % (-self.id,)
        return 'abcdefghijklmnopqrstuvwxyz'[-(self.id+1)]

class arrow (_type):
    def __init__ (self, t0, t1):
        self.t0 = t0
        self.t1 = t1

    def __repr__ (self):
        return '%r->%r' % (self.t0, self.t1)
    
# constraints

def constraint_repr (kind, args):
    if args:
        return '[%s %s]' % (kind, ' '.join ([repr(x) for x in args]))
    else:
        return '[%s]' % (kind,)

class constraint:
    kind = 'abstract'
    args = ()

    def __repr__ (self):
        return constraint_repr (self.kind, self.args)

class c_true (constraint):
    kind = 'true'
    args = ()

class c_false (constraint):
    kind = 'true'
    args = ()

class c_equals (constraint):
    kind = 'equals'
    def __init__ (self, *args):
        self.args = args
    def __repr__ (self):
        return constraint_repr ('=', self.args)

class c_and (constraint):
    kind = 'and'
    def __init__ (self, c0, c1):
        self.args = (c0, c1)
        
class c_exists (constraint):
    kind = 'exists'
    def __init__ (self, vars, sub):
        self.args = (vars, sub)
        self.vars = vars
        self.sub = sub

class c_is (constraint):
    # <x> has type <t> iff <t> is an instance of the type scheme associated with <x>
    kind = 'is'
    def __init__ (self, x, t):
        self.args = (x, t)
        self.x = x
        self.t = t

class c_let (constraint):
    kind = 'let'
    def __init__ (self, formal, init, body):
        self.args = (formal, init, body)
        self.formal = formal
        self.init = init # this is poorly named, should be 'type' or 'scheme'
        self.body = body

# XXX consider calling this a 'scheme' or 'typescheme'?
class c_forall (constraint):
    kind = 'forall'
    def __init__ (self, vars, constraint, type):
        self.args = (vars, constraint, type)
        self.vars = vars
        self.constraint = constraint
        self.type = type

# stack frames
class frame:
    kind = 'abstract'
    args = ()

class s_empty (frame):
    kind = 'empty'

empty = s_empty()

class s_and (frame):
    def __init__ (self, c):
        self.args = (c,)
        self.constraint = c

class s_exists (frame):
    def __init__ (self, vars):
        self.args = (vars,)
        self.vars = vars

class s_let (frame):
    def __init__ (self, formal, vars, type, body):
        self.args = (formal, vars, type, body)
        self.formal = formal
        self.vars = vars
        self.type = type
        self.body = body

class s_env (frame):
    def __init__ (self, formal, type):
        self.args = (formal, type)
        self.formal = formal
        self.type = type

# this is a two-phase algorithm
# 1) constraint generation
# 2) constraint solving

# original version without the let construct

class typer:

    def __init__ (self, verbose=False):
        pass

    def go (self, exp):
        env = {}
        t = tvar()
        x = self.generate (env, exp, t)
        return x

    def generate (self, env, exp, t):
        if exp.is_a ('varref'):
            return c_equals (env[exp.name], t)
        elif exp.is_a ('function'):
            a1 = tvar()
            a2 = tvar()
            # a dict will suffice because we've alpha-converted <exp>
            env[exp.formals[0].name] = a1
            sub0 = self.generate (env, exp.body, a2)
            sub1 = c_equals (t, arrow (a1, a2))
            return c_exists ((a1, a2), c_and (sub0, sub1))
        elif exp.is_a ('application'):
            e1 = exp.rator
            e2 = exp.rands[0]
            a = tvar()
            sub0 = self.generate (env, e1, arrow (a, t))
            sub1 = self.generate (env, e2, a)
            return c_exists ((a,), c_and (sub0, sub1))
        elif exp.is_a ('literal'):
            return c_equals (exp.type, t)
        else:
            raise ValueError

# variant with def/let construct

class typer:

    # Note: in this version, there is no environment (or rather, it's
    #   embedded in the constraint via the def/let mechanism).

    def go (self, exp):
        t = tvar()
        x = self.generate (exp, t)
        return x

    def generate (self, exp, t):
        if exp.is_a ('varref'):
            return c_is (exp.name, t)
        elif exp.is_a ('function'):
            assert (len (exp.formals) == 1)
            a1 = tvar()
            a2 = tvar()
            sub0 = self.generate (exp.body, a2)
            let0 = c_let (exp.formals[0].name, a1, sub0)
            # XXX: in ATTPL, this is a c_supertype relation
            sub1 = c_equals (t, arrow (a1, a2))
            return c_exists ((a1, a2), c_and (let0, sub1))
        elif exp.is_a ('application'):
            e1 = exp.rator
            e2 = exp.rands[0]
            a = tvar()
            sub0 = self.generate (e1, arrow (a, t))
            sub1 = self.generate (e2, a)
            return c_exists ((a,), c_and (sub0, sub1))
        elif exp.is_a ('let_splat'):
            # let_splat ::= names, inits, body
            x = tvar()
            assert (len(exp.names) == 1)
            init0 = self.generate (exp.inits[0], x)
            body0 = self.generate (exp.body, t)
            return c_let (exp.names[0].name, c_forall ((x,), init0, x), body0)
        elif exp.is_a ('literal'):
            return c_equals (exp.type, t)
        else:
            raise ValueError

class UnboundVariable (Exception):
    pass

class Continue (Exception):
    pass

# I believe the 'union-find' data-structure/algorithm is another way
#   of describing HM's use of type variable lookup with path
#   compression.

class solver:

    def __init__ (self):
        pass

    def solve (self, c):

        exists = []
        s, u, c = [], c_true(), c

        def push (x):
            s.insert (0, x)

        def pop ():
            s.pop (0)

        while 1:
            # "rules s-ex-1 to s-ex-4 float existential quantifiers out of
            # the unification problem into the stack and through the stack
            # up to the nearest enclosing let frame, if there is any, or
            # to the outermost level, otherwise."

            # In the mini-prototype code, exists & forall are both encoded
            # as CLet forms.

            print 'S:',
            pprint_stack (s)
            print
            print 'U:', u
            print 'C:', c

            print 'ftv(U)=', ftv (u)

            # easy access to the top two elements of the stack
            if len(s) > 0:
                s0 = s[0]
            else:
                s0 = s_empty()
            if len(s) > 1:
                s1 = s[1]
            else:
                s1 = s_empty()

            # --- unification ---
            unified = True

            if is_a (u, _type):
                print 's-single'
                u = c_true()
            elif u.kind is 'and' and is_a (u.args[0], c_true):
                print 's-true'
                u = u.args[1]
            elif u.kind is 'and' and is_a (u.args[1], c_true):
                print 's-true'
                u = u.args[0]
            else:
                # s-fuse is probably next
                # XXX rest of the unifier here
                unified = False
            
            if unified:
                print 'unification - U:', u

            print '-----------------------------'

            #trace()
            raw_input()
            # --- solver ---            

            # XXX everywhere here, alpha-conversion to avoid capture
            if is_a (u, c_exists):
                print 's-ex-1'
                push (s_exists (u.vars))
                u = u.sub
            elif is_a (s0, s_and) and is_a (s1, s_exists):
                print 's-ex-2'
                # and D              exists X 
                # exists X   ==>     and D    
                # S'                 S'       
                s[0], s[1] = s[1], s[0]
            elif is_a (s0, s_let) and is_a (s1, s_exists):
                print 's-ex-3'
                # nuke the <exists> at s1, merge its vars with s0
                pop(); pop()
                push (s_let (s0.formal, s0.vars + s1.vars, s0.type, s0.body))
            elif is_a (s0, s_env) and is_a (s1, s_exists):
                print 's-ex-4'
                # env;exists;... => exists;env;...
                pop(); pop()
                push (s0)
                push (s1)
            elif is_a (s0, s_exists):
                print 's-discard-ex'
                exists.extend (s0.vars)
                pop()
            elif is_a (c, c_equals):
                print 's-solve-eq'
                s, u, c = s, c_and (u, c), c_true()
            elif is_a (c, c_is) and is_a (c.x, str):
                print 's-solve-id'
                scheme, type = self.lookup (c.x, s), c.t
                # assert that scheme.type is a tvar
                # "Recall that if <sigma> is of the form
                # forall(X0..XN[U].X) where X0..XN#ftv(T), then c_is(sigma, T)
                # stands for exists(X0..XN.(U ^ X=T)."
                print 'scheme=', scheme
                print 'type=', type
                c = c_exists (scheme.vars, c_and (scheme.constraint, c_equals (scheme.type, type)))
            elif is_a (c, c_and):
                print 's-solve-and'
                push (s_and (c.args[0]))
                c = c.args[1]
            elif is_a (c, c_exists):
                print 's-solve-ex'
                # XXX alpha conversion
                push (s_exists (c.vars))
                c = c.sub
            elif is_a (c, c_let) and is_a (c.init, c_forall) and is_a (c.init.constraint, c_exists):
                # try to stop s-solve-let from floating the exists outside of the let
                print 's-ex-3*'
                c = c_let (c.formal, c_forall (c.init.vars + c.init.constraint.vars, c.init.constraint.sub, c.init.type), c.body)
            elif is_a (c, c_let):
                print 's-solve-let'
                if is_a (c.init, c_forall):
                    push (s_let (c.formal, c.init.vars, c.init.type, c.body))
                    c = c.init.constraint
                else:
                    # let x: T in C == let x: forall/0.T in C
                    push (s_let (c.formal, (), c.init, c.body))
                    c = c_true()
            elif is_a (c, c_true):
                if is_a (s0, s_and):
                    print 's-pop-and'
                    pop()
                    c = s0.constraint
                elif is_a (s0, s_let) and not is_a (s0.type, tvar):
                    print 's-name-2'
                    x = tvar()
                    pop()
                    push (s_let (s0.formal, s0.vars + (x,), x, s0.body))
                elif is_a (s0, s_let):
                    # s-compress, s-unname, s-letall, s-pop-let
                    # these rules all require an examination of the variables
                    #  bound by this <let>.
                    try:
                        ftvu = ftv (u)
                        for var in s0.vars:
                            for e in iterate_conj (u):
                                tv, ty = self.collect_tvars (e)
                                if var in tv and len(tv) >= 2:
                                    # for s-compress, we want to know if there's a multi-equation
                                    # of the form Y=Z=e (where Y is bound by this <let>)
                                    print 's-compress'
                                    # subst will leave this one in the form Z=Z=e, which
                                    #   should be caught by s-stutter in the unifier.
                                    Y = var
                                    if tv[0] is not var:
                                        Z = tv[0]
                                    else:
                                        Z = tv[1]
                                    u = c_and (Y, self.subst (Z, Y, u))
                                    if s0.type == var:
                                        # XXX hack
                                        s0.type = Z
                                    raise Continue
                                elif var in tv and len(tv) == 1 and ftvu.count (var) == 1 and s0.type != var:
                                    # for s-unname, we want to know if Y is mentioned only once
                                    # in U, if so, we can nix it.
                                    print 's-unname'
                                    vars = s0.vars[:]
                                    vars.remove (var)
                                    pop()
                                    # remove Y from the <let>
                                    push (s_let (s0.formal, vars, s0.type, s0.body))
                                    # remove Y from <u> - since we know it's only in this
                                    #  one multi-equation, I'm gonna cheat and just mutate it.
                                    e.args = [x for x in e.args if x is not Y]
                                    raise Continue
                        #elif s-letall:
                        else:
                            # pop-let is a fall-through - after all
                            # the above conditions have been met it turns the <let>
                            # into an <env>.
                            print 's-pop-let'
                            u1, u2 = self.split_u (s0, u)
                            pop()
                            push (s_env (s0.formal, c_forall (s0.vars, u2, s0.type)))
                            u = u1
                            c = s0.body
                    except Continue:
                        pass
                elif is_a (s0, s_env):
                    print 's-pop-env'
                    pop()
                elif is_a (s0, s_empty):
                    return u
                else:
                    raise ValueError ("unexpected")
            else:
                print 'no rule applies'
                trace()
                
    def split_u (self, s0, u):
        # u1 is made up entirely of 'old' variables - i.e., not bound by
        #  the <let> in s0.
        u1 = []
        u2 = []
        young = set (s0.vars)
        for e in iterate_conj (u):
            flag = True
            for var in ftv (e):
                # only if all of them are old can it go in u1
                if var in young:
                    flag = False
            if flag:
                u1.append (e)
            else:
                u2.append (e)
        return list_to_conj (u1), list_to_conj (u2)

    def lookup (self, x, s):
        # XXX side conditions
        for f in s:
            if is_a (f, s_and):
                continue
            elif is_a (f, s_exists):
                continue
            elif is_a (f, s_let):
                if f.formal == x:
                    raise ValueError ("shouldn't happen?")
                continue
            elif is_a (f, s_env):
                if f.formal != x:
                    continue
                else:
                    return f.type
        else:
            raise UnboundVariable (x)

    def collect_tvars (self, e):
        # split a multi-equation into tvars and types
        tv = []
        ty = []
        for e0 in e.args:
            if is_a (e0, tvar):
                tv.append (e0)
            else:
                ty.append (e0)
        return tv, ty

    # need a 'map_constraint' that will map a function over any constraint.
    def subst (self, v0, v1, e):
        print ' -- subst before:', e
        def f (e):
            if is_a (e, tvar):
                if e.id == v0.id:
                    return v1
                else:
                    return e
            elif is_a (e, arrow):
                return arrow (f(e.t0), f(e.t1))
            if is_a (e, c_equals):
                return c_equals (*[f(x) for x in e.args])
            elif is_a (e, c_and):
                return c_and (*[f(x) for x in e.args])
            else:
                trace()
        r = f (e)
        print ' -- subst after :', r
        return r

def ftv (e):
    r = []
    def recur (bound, e):
        if is_a (e, c_equals):
            for arg in e.args:
                recur (bound, arg)
        elif is_a (e, c_and):
            for arg in e.args:
                recur (bound, arg)
        elif is_a (e, c_exists):
            recur (bound.union (e.vars), e.sub)
        elif is_a (e, c_is):
            recur (bound, e.t)
        elif is_a (e, c_let):
            recur (bound, e.init)
            recur (bound, e.body)
        elif is_a (e, c_forall):
            bound = bound.union (e.vars)
            recur (bound, e.constraint)
            recur (bound, e.type)
        elif is_a (e, tvar):
            r.append (e)
        elif is_a (e, arrow):
            recur (bound, e.t0)
            recur (bound, e.t1)
        elif is_a (e, c_true):
            pass
        elif is_a (e, c_false):
            pass
        else:
            raise NotImplementedError
    recur (set(), e)
    return r

def iterate_conj (u):
    if is_a (u, c_and):
        for x in iterate_conj (u.args[0]):
            yield x
        for x in iterate_conj (u.args[1]):
            yield x
    else:
        yield u

def list_to_conj (l):
    if len(l) == 0:
        return c_true()
    elif len(l) == 1:
        return c_and (c_true(), l[0])
    else:
        r = l[0]
        for x in l[1:]:
            r = c_and (r, x)
        return r

def pprint_stack (s):

    W = sys.stdout.write

    W ('\n')
    n = len(s)
    for i in range (n):
        W ('%2d: ' % (i,))
        si = s[i]
        if is_a (si, s_empty):
            W ('[]')
        elif is_a (si, s_and):
            W ('[] ^ %s' % si.args)
        elif is_a (si, s_exists):
            W ('exists%r.[]' % si.args)
        elif is_a (si, s_let):
            W ('let %s: forall%r[[]].%r in %r' % si.args)
        elif is_a (si, s_env):
            W ('env %s: %r in []' % si.args)
        else:
            raise NotImplementedError
        W ('\n')

def read_string (s):
    import cStringIO
    import lisp_reader
    sf = cStringIO.StringIO (s)
    r = lisp_reader.reader (sf)
    return r.read()

def test (s):
    import transform
    import lambda_tree
    from pprint import pprint as pp
    global tvar_counter
    tvar_counter = -1
    # wrap everything in a top-level <let>
    s = "(let ((top %s)) top)" % s
    exp = read_string (s)
    t = transform.transformer (1)
    exp2 = t.go ([exp])
    w = lambda_tree.walker()
    exp3 = w.go (exp2)
    # alpha conversion
    var_dict = lambda_tree.rename_variables (exp3)
    t = typer()
    constraints = t.go (exp3)
    pp (constraints)
    u = solver().solve (constraints)
    print u

test ("(lambda (x) x)")
#test ("(lambda (x) (lambda (y) x))")
#test ("(let ((f (lambda (x) x))) f)")
#test ("((lambda (x) x) (lambda (x) x))")

if False:
    # for the simple variant:
    test ("(lambda (x) x)")
    test ("((lambda (x) x) (lambda (x) x))")
    test ("((lambda (x) x) (lambda (y) y))")
    test ("((lambda (x) x) 3)")
    # for the def/let variant:
    test ("(let ((x 3)) (lambda (y) (y x)))")
